#include 	<X11/Xlib.h>
#include 	<X11/Xutil.h>
#include    <X11/StringDefs.h>
#include    <X11/cursorfont.h>
#include    <X11/Intrinsic.h>
#include 	<stdio.h>
#include 	<stdlib.h>
#include 	<string.h>
#include 	<math.h>
#include 	<fcntl.h>
#include 	<sys/types.h>
#include 	<unistd.h>

#include 	"typedef.h"
#include 	"drawchar.h"
#include 	"sintbl.h"

/**********************************************************************/
/*                                                                    */
/*  DRAWCHAR.C  - Outline Font Rasterizer & Primitive Font Manager    */
/*                                                                    */
/*  Programmed by : Lee, Jun-Hee  (member of Technical staff, HCI)    */
/*  Modified By Kim Bum Chul 1995.12.22				      */
/*                                                                    */
/*                                                                    */
/*  Copyright (c) 1993 by Human Computers,Inc.                        */
/*  Copyright (c) 1995 by TriGem Microsystems, Inc.                   */
/*                                                                    */
/**********************************************************************/

/*
 * set tab size = 4 to view this source
 */

typedef struct _FONTINFO {
	char	FontName[32];		  /* typeface name */
	WORD	Version;			  /* version # */
	WORD	CharSet;              /* Hangul or English */
	WORD	CodingSystem;         /* English, KS, Johap, etc */
	WORD	Pitch;                /* fixed or variable */
	WORD	StartCode;            /* starting code (if KS, starting row #) */
	WORD    CharNum;              /* numer of character */
	WORD	RowNum;				  /* row # if KS code */
	WORD    EmSize; 		      /* design em size  */
	WORD    DefaultWidth;         /* default width (relative to DesignMeshSize)  */
	LONG    MapOffset;            /* map table offset  */
	LONG    WidthTblOffset;       /* width table offset  */
	LONG    GlyphOffset;          /* glyph data offset  */
	char	FileName[128];		  /* actual filename */
	SHORT	*WidthTable;		  /* width table in memory */
	char	Reserved[48];
} FONTINFO, * LPFONTINFO;


/*  Pitch */

#define		FONT_FIXED		0
#define		FONT_VARIABLE	1

/* CodingSystem */

#define		FONT_ENG		0
#define		FONT_KSCODE		100
#define		FONT_JOHAP		200

/* CharSet */

#define		FONT_LATIN		0
#define		FONT_HANGUL		1
#define		FONT_HANJA		2
#define		FONT_KSSYMBOL	3

/* Map Entry */

typedef struct _MAPENTRY {
	SHORT	nGlyphSize;
	LONG	lOffset;
} MAPENTRY;


/* private definitions */

#define		MAX_FONTDATA	4096
#define		MAX_BITMAPSIZE	512
#define		MAX_BITMAPBUF	45000

#define		MAGIC_PROP		0xAA

#define		MOVETO			0x10
#define		LINETO			0x20
#define		CURVETO			0x40
#define		ENDCHAR			0x80


typedef struct _BITMAPMETRIC {
			SHORT	Width;
			SHORT	Height;
			SHORT	xOffset;
			SHORT	yOffset;
		} BITMAPMETRIC, *LPBITMAPMETRIC;

#define		MAX_HANGULFONT		1
#define		MAX_HANJAFONT		1
#define		MAX_SPECFONT		1
#define		MAX_ENGFONT			1

static FONTINFO			HangulFontInfo[MAX_HANGULFONT];
static FONTINFO			HanjaFontInfo[MAX_HANJAFONT];
static FONTINFO			SpecFontInfo[MAX_SPECFONT];
static FONTINFO			EngFontInfo[MAX_ENGFONT];
static LPFONTINFO		lpCurHangulFont, lpCurEngFont, lpCurHanjaFont, lpCurSpecFont;

static char		FontDir[100];

static LPSTR	lpFontData, lpCurFontData;
static LPSTR	lpBitmapBuffer;
static short	nEmSize = 256;
static short	nMaxSubdivDepth = 3;
static short	nCurFontSizeX, nCurFontSizeY;
static short	nCharBitmapWidth, nCharBitmapHeight, nCharXBytes;
static LONG		xShift, yShift;
static LPSTR	lpCurFrameBuf;

typedef struct _MATRIX {
			LONG	a, b, c, d, tx, ty;
		} MATRIX;

static MATRIX	CTM;


static SHORT	*alpPoint[MAX_BITMAPSIZE * 2];
static BYTE		anCount[MAX_BITMAPSIZE * 2];


#define		ABS(x)	(((x) > 0) ? (x) : -(x))

/*static BYTE mask[8]  = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
*/
static BYTE mask[8]  = { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };

/*static BYTE mask[8]  = { 0x10, 0x20, 0x40, 0x80, 0x01, 0x02, 0x04, 0x08 };
*/

/**********************************************************************/

static BOOL		OpenFontFile();
static BOOL		InitBuffers();
static void		FreeBuffers();
static BOOL		_DrawChar();
static void		ScaleCoord();
static BOOL		GetBitmapFromOutline();
static BOOL		CalcCharMetric();
static void		TogglePixel();
static void		FillHorizLine();
static void		DrawLine();
static void		DrawBezier();
static SHORT	GetWord();
static BOOL		DrawGlyph();
static void		Sort();
static void		Fill();

static UBCHAR	KSSM2KS();
void			swap2bytes();
void			swap4bytes();

/***********************************************************************/


char	*HangulFontName[MAX_HANGULFONT] = {
			"hanmoon.fnt"
		};
char	*HanjaFontName[MAX_HANJAFONT]= {
			"hjmj.fnt"
		};
char	*SpecFontName[MAX_SPECFONT] = {
			"sym.fnt",
		};
char	*EngFontName[MAX_ENGFONT] = {
			"english.fnt"
		};

/*********************************************************************/

BOOL InitFontMachinery(FontPath)
char *FontPath;
{
	short		i;

	if (!InitBuffers()) return FALSE;
	ResetCTM();
	nCurFontSizeX = 1;
	nCurFontSizeY = 1;
	strcpy(FontDir, FontPath);

	for (i = 0; i < MAX_HANGULFONT; i++)
		OpenFontFile(&HangulFontInfo[i], HangulFontName[i]);
	for (i = 0; i < MAX_HANJAFONT; i++)
		OpenFontFile(&HanjaFontInfo[i], HangulFontName[i]);
	for (i = 0; i < MAX_SPECFONT; i++)
		OpenFontFile(&SpecFontInfo[i], SpecFontName[i]);
	for (i = 0; i < MAX_ENGFONT; i++)
		OpenFontFile(&EngFontInfo[i], EngFontName[i]);

	lpCurHangulFont = &HangulFontInfo[0];
	lpCurHanjaFont = &HanjaFontInfo[0];
	lpCurSpecFont = &SpecFontInfo[0];
	lpCurEngFont = &EngFontInfo[0];

	return TRUE;
}


static BOOL OpenFontFile(fi, filename)
LPFONTINFO fi;
char *filename;
{
	int		fd;
	char	s[100];
	unsigned 	len;
	long 	curp;

	sprintf(s, "%s/%s", FontDir, filename);
 	fd = open(s, O_RDONLY);
 	if (fd == -1) goto err;
 	if (read(fd, fi, sizeof(FONTINFO)) != sizeof(FONTINFO)) goto err; 
 
/*
	swap2bytes(&fi->Version);
	swap2bytes(&fi->CharSet);
	swap2bytes(&fi->CodingSystem);
	swap2bytes(&fi->Pitch);
	swap2bytes(&fi->StartCode);
	swap2bytes(&fi->CharNum);
	swap2bytes(&fi->RowNum);
	swap2bytes(&fi->EmSize);
	swap2bytes(&fi->DefaultWidth);
*/
	/*swap2bytes((unsigned short *)&fi->MapOffset);
	swap2bytes((unsigned short *)&fi->WidthTblOffset);
	swap2bytes((unsigned short *)&fi->GlyphOffset);*/

	if (fi->CodingSystem == 100) {
		fi->MapOffset = 256L;
		fi->WidthTblOffset = 0L;
		fi->GlyphOffset = 14356L;
	} else if (fi->CodingSystem == 0) {
		fi->MapOffset = 256L;
		fi->WidthTblOffset = 1024L;
		fi->GlyphOffset = 1280L;
	}

	if (fi->Pitch == FONT_VARIABLE) {
		len = fi->CharNum * sizeof(SHORT);
		fi->WidthTable = (SHORT *)malloc(len);
		if (fi->WidthTable == NULL) goto err;
		lseek(fd, fi->WidthTblOffset, 0);
		if (read(fd, fi->WidthTable, len) != len) goto err;
	}

	close(fd);
	strcpy(fi->FileName, s);
	return TRUE;

err:
	close(fd);
	return FALSE;
}


void CloseFontMachinery()
{
	FreeBuffers();
}

static BOOL InitBuffers()
{
	short		i;

	lpFontData = (LPSTR)malloc(MAX_FONTDATA);
	lpBitmapBuffer = (LPSTR)malloc(MAX_BITMAPBUF);
	if (!lpFontData || !lpBitmapBuffer) return FALSE;

	for (i = 0; i < MAX_BITMAPSIZE * 2; i++) {
		alpPoint[i] = (SHORT *)malloc(30 * sizeof(SHORT));
		if (alpPoint[i] == NULL) return FALSE;
	}

	return TRUE;
}

static void FreeBuffers()
{
	short		i;

	free(lpFontData);
	free(lpBitmapBuffer);
	for (i = 0; i < MAX_BITMAPSIZE * 2; i++)
		free(alpPoint[i]);
}

void ResetCTM()
{
	CTM.a = 1024;
	CTM.b = 0;
	CTM.c = 0;
	CTM.d = 1024;
	CTM.tx = 0;
	CTM.ty = 0;
}

void SetFontSize(xSize, ySize)
int xSize, ySize;
{
	nCurFontSizeX = xSize;
	nCurFontSizeY = ySize;

	if (xSize > 128 || ySize > 128)
		nMaxSubdivDepth = 4;
	else
		nMaxSubdivDepth = 3;
}


void RotateFont(Angle)
int		Angle;
{
	LONG	SinVal, CosVal;
	LONG	a, b, c, d, tx, ty;

	Angle %= 360;
	if (Angle < 0) Angle += 360;
	if (Angle == 0) return;

	CosVal = CosTbl[Angle];
	SinVal = SinTbl[Angle];

	a = CTM.a * CosVal - CTM.b * SinVal;
	b = CTM.a * SinVal + CTM.b * CosVal;
	c = CTM.c * CosVal - CTM.d * SinVal;
	d = CTM.c * SinVal + CTM.d * CosVal;
	tx = CTM.tx * CosVal - CTM.ty * SinVal;
	ty = CTM.tx * SinVal + CTM.ty * CosVal;

	CTM.a = a >> 10;
	CTM.b = b >> 10;
	CTM.c = c >> 10;
	CTM.d = d >> 10;
	CTM.tx = tx >> 10;
	CTM.ty = ty >> 10;
}

void SlantFont(Angle)
int		Angle;
{
	LONG	SinVal, CosVal;

	Angle %= 360;
	if (Angle >= 90 || Angle <= -90) return;
	if (Angle < 0) Angle += 360;
	if (Angle == 0) return;

	SinVal = SinTbl[Angle];
	CosVal = CosTbl[Angle];

	CTM.a -= CTM.c * SinVal / CosVal;
	CTM.b -= CTM.d * SinVal / CosVal;
	CTM.tx -= CTM.ty * SinVal / CosVal;
}

void SetFont(CodeArea, FontId)
int CodeArea, FontId;
{
	switch (CodeArea) {
		case ENGLISH :
			if (FontId >= 0 && FontId < MAX_ENGFONT)
				lpCurEngFont = &EngFontInfo[FontId];
			break;
		case HANGUL	 :
			if (FontId >= 0 && FontId < MAX_HANGULFONT)
				lpCurHangulFont = &HangulFontInfo[FontId];
			break;
		case HANJA	 :
			if (FontId >= 0 && FontId < MAX_HANJAFONT)
				lpCurHanjaFont = &HanjaFontInfo[FontId];
			break;
		case SPECIAL :
			if (FontId >= 0 && FontId < MAX_SPECFONT)
				lpCurSpecFont = &SpecFontInfo[FontId];
			break;
	}
}

void DrawString(display, d, gchan, x, y, str)
Display         *display;
Drawable        d;
int x, y;
unsigned char *str;
{
	int		xinc, yinc;
	int c;

	while (*str) {
		if (*str < 128) {
			c=*str++;
		}
		else {
                        c = (*str << 8) + *(str+1);
                        str += 2;
		}
                DrawChar(display, d, gchan, x, y, c, &xinc, &yinc);
		x += xinc;
		y += yinc;
	}
}


void DrawChar(display, d, gchan, x, y, ch, xInc, yInc)
Display         *display;
Drawable        d;
GC gchan;
int x, y;
int ch;
int *xInc, *yInc;
{
	UBCHAR		ch2;
	int			row, fd;
	short		nWidth;
	LPFONTINFO	fi;
	int			index;

	if (ch < 0x80) fi = lpCurEngFont;
	else if (ch >= 0x8440 && ch <= 0xd3fe) fi = lpCurHangulFont;
	else if (ch >= 0xd931 && ch <= 0xdefe) fi = lpCurSpecFont;
	else if (ch >= 0xe031 && ch <= 0xf9fe) fi = lpCurHanjaFont;
	else goto err;

	if ((fd = open(fi->FileName, O_RDONLY)) == -1) goto err;
	switch (fi->CodingSystem) {

		case FONT_ENG:
			if (ch < fi->StartCode || ch >= fi->StartCode + fi->CharNum)
				goto err;
			index = ch - fi->StartCode;
			_DrawChar(display, d, gchan, x, y, fi, fd, index);
			break;

		case FONT_KSCODE:
			if ((ch2 = ch) == 0xffff) goto err;
			row = (ch2 >> 8) - fi->StartCode;
			if (row < 0 || row >= fi->RowNum) goto err;
			index = row * 94 + (ch2 & 0x00ff) - 0xa1;
			_DrawChar(display, d, gchan, x, y, fi, fd, index);
			break;

		default:
			goto err;
	}
	close(fd);

	if (fi->Pitch == FONT_FIXED)
		nWidth = fi->DefaultWidth;
	else {
		nWidth = fi->WidthTable[index];
/*
		swap2bytes(&nWidth);
*/
	}

	GetCharNext(nWidth, xInc, yInc);
	return;

err:
	*xInc = *yInc = 0;
	close(fd);
	return;
}

short GetCharWidth(ch)
UBCHAR ch;
{
	UBCHAR		ch2;
	LPFONTINFO	fi;
	int			row, index;

	if (ch < 0x80) fi = lpCurEngFont;
	else if (ch >= 0x8440 && ch <= 0xd3fe) fi = lpCurHangulFont;
	else if (ch >= 0xd931 && ch <= 0xdefe) fi = lpCurSpecFont;
	else if (ch >= 0xe031 && ch <= 0xf9fe) fi = lpCurHanjaFont;
	else goto err;

	if (fi->Pitch == FONT_FIXED) {
		return fi->DefaultWidth;
	}
	else {
		switch (fi->CodingSystem) {
			case FONT_ENG:
				if (ch < fi->StartCode || ch >= fi->StartCode + fi->CharNum) return -1;
				index = ch - fi->StartCode;
				break;
			case FONT_KSCODE:
				if ((ch2 = ch) == 0xffff) return -1;
				row = (ch2 >> 8) - fi->StartCode;
				if (row < 0 || row >= fi->RowNum) return -1;
				index = row * 94 + (ch2 & 0x00ff) - 0xa1;
				break;
			default:
				return -1;
		}
		return fi->WidthTable[index];
	}
err:
	return -1;
}


static BOOL _DrawChar(display, d, gchan, x, y, fi, fd, index)
Display         *display;
Drawable        d;
GC gchan;
int x, y;
LPFONTINFO fi;
int fd, index;
{
	BITMAPMETRIC	Met;
	int				nRet;
	MAPENTRY		me;

	Pixmap			pmaphan;

	Display			*disp;
	Window			win;

	
	lseek(fd, fi->MapOffset + (LONG)index * (sizeof(MAPENTRY)-2), 0);
/*	if (read(fd, &me, sizeof(MAPENTRY)-2) != sizeof(MAPENTRY)-2) return 0;
*/

	read(fd,&me.nGlyphSize,2);
	read(fd,&me.lOffset,4);
/*
	swap2bytes(&me.nGlyphSize);
	swap4bytes(&me.lOffset);
*/

	lseek(fd, me.lOffset, 0);
	if (read(fd, lpFontData, me.nGlyphSize) != me.nGlyphSize) return 0;
	if (me.nGlyphSize == 0) lpFontData[0] = ENDCHAR;

	nRet = GetBitmapFromOutline(lpFontData, lpBitmapBuffer, &Met);
	if (!nRet) return 0;

	x += Met.xOffset;
	y -= (Met.yOffset + Met.Height);
/**************************************************************************************
 *
 *
 */
	if (check_size == 1) {
		check_x = x;
		check_y	= y;
		check_width = Met.Width;
		return 1;
	}	
    pmaphan = XCreateBitmapFromData(display, d, lpBitmapBuffer, Met.Width, Met.Height);

    XSetStipple(display, gchan, pmaphan);
	XSetFillStyle(display, gchan, FillStippled);
	XSetTSOrigin(display, gchan, x, y);
    XFillRectangle(display, d, gchan, x, y, Met.Width, Met.Height);


	XFreePixmap(display, pmaphan);

	XSetFillStyle(display, gchan, FillSolid);

/*    XDrawRectangle(disp, win, gchan, x, y, Met.Width, Met.Height);
*/
/*	XFlush(disp);
*/

/**************************************************************************************/

	return 1;
}

void GetCharNext(nCharWidth, dx, dy)
int nCharWidth, *dx, *dy;
{
	LONG	DX, DY;

	DX = (LONG)nCharWidth * nCurFontSizeX / nEmSize;
	DY = 0;
	*dx = (DX * CTM.a + DY * CTM.b + CTM.tx) >> 10;
	*dy = - (DX * CTM.c + DY * CTM.d + CTM.ty) >> 10;
}

static void ScaleCoord(X, Y)
LONG *X, *Y;
{
	LONG	XX = *X * (nCurFontSizeX * 1024L / nEmSize);
	LONG	YY = *Y * (nCurFontSizeY * 1024L / nEmSize);

	*X = (XX * CTM.a + YY * CTM.b + CTM.tx) >> 10;
	*Y = (XX * CTM.c + YY * CTM.d + CTM.ty) >> 10;
}

static short GetBitmapFromOutline(lpFontDataBuf, lpBitmapBuf, lpMet)
LPSTR lpFontDataBuf, lpBitmapBuf;
LPBITMAPMETRIC lpMet;
{
	lpCurFontData = lpFontDataBuf;

	if (!CalcCharMetric(lpMet))
		return 0;

	nCharBitmapWidth = lpMet->Width;
	nCharBitmapHeight = lpMet->Height;
	nCharXBytes = (nCharBitmapWidth + 7) / 8;
	lpCurFrameBuf = lpBitmapBuf;

	if ((LONG)nCharXBytes * lpMet->Height < MAX_BITMAPBUF) {
		memset(lpCurFrameBuf, 0, (size_t)(nCharXBytes * nCharBitmapHeight));
		if (!DrawGlyph()) return 0;
		Fill();
		return 1;
	}
	return 0;
}


static BOOL CalcCharMetric(lpMet)
LPBITMAPMETRIC lpMet;
{
	short		x11, y11, x22, y22;
	LONG	xUpLeft, yUpLeft, xUpRight, yUpRight;
	LONG	xDownLeft, yDownLeft, xDownRight, yDownRight;
	LONG	xMin, yMin, xMax, yMax;

	if (*lpCurFontData == MAGIC_PROP) {
		lpCurFontData++;
		x11 = GetWord() - 4;
		y11 = GetWord() - 4;
		x22 = GetWord() + 4;
		y22 = GetWord() + 4;
		if (x11 >= x22 || y11 >= y22)
			return FALSE;
	}
	else {
		x11 = y11 = 0;
		x22 = y22 = nEmSize;
	}

	xUpLeft = xDownLeft = x11;
	xUpRight = xDownRight = x22;
	yUpLeft = yUpRight = y11;
	yDownLeft = yDownRight = y22;

	ScaleCoord(&xUpLeft, &yUpLeft);
	ScaleCoord(&xUpRight, &yUpRight);
	ScaleCoord(&xDownLeft, &yDownLeft);
	ScaleCoord(&xDownRight, &yDownRight);

	xMin = xUpLeft;
	if (xMin > xUpRight) xMin = xUpRight;
	if (xMin > xDownLeft) xMin = xDownLeft;
	if (xMin > xDownRight) xMin = xDownRight;
	xMax = xUpLeft;
	if (xMax < xUpRight) xMax = xUpRight;
	if (xMax < xDownLeft) xMax = xDownLeft;
	if (xMax < xDownRight) xMax = xDownRight;
	yMin = yUpLeft;
	if (yMin > yUpRight) yMin = yUpRight;
	if (yMin > yDownLeft) yMin = yDownLeft;
	if (yMin > yDownRight) yMin = yDownRight;
	yMax = yUpLeft;
	if (yMax < yUpRight) yMax = yUpRight;
	if (yMax < yDownLeft) yMax = yDownLeft;
	if (yMax < yDownRight) yMax = yDownRight;

	lpMet->Width = ((xMax - 1) >> 10) - (xMin >> 10) + 1;
	lpMet->Height = ((yMax - 1) >> 10) - (yMin >> 10) + 1;
	lpMet->xOffset = (short)(xMin >> 10);
	lpMet->yOffset = (short)(yMin >> 10);

	xShift = (LONG)lpMet->xOffset << 10;
	yShift = (LONG)lpMet->yOffset << 10;

	return TRUE;
}


static void TogglePixel(x, y)
int		x, y;
{
	if (x < 0 || x >= nCharBitmapWidth || y < 0 || y >= nCharBitmapHeight)
		return;
	alpPoint[y][anCount[y]++] = x;
}

static void FillHorizLine(x1, x2, y)
int		x1, x2, y;
{
	LPSTR	p;
	int		t, k;

	if (x1 > x2) {
		t = x1; x1 = x2; x2 = t;
	}
	p = lpCurFrameBuf + (y * nCharXBytes);
	for (k = x1; k <= x2; k++)
		*(p + (k >> 3)) |= mask[k & 7];
}


static void DrawLine(x1, y1, x2, y2)
LONG		x1, y1, x2, y2;
{
	int		x, y;
	int		iy1, iy2, starty, endy, incdec;
	LONG	xx, dx, dy;

	x1 = (x1 + 16) >> 5;
	y1 = (y1 + 16) >> 5;
 	x2 = (x2 + 16) >> 5;
 	y2 = (y2 + 16) >> 5;

 	iy1 = (int)((y1 + 16) >> 5);
 	iy2 = (int)((y2 + 16) >> 5);
 
 	if (iy1 == iy2) {
 		FillHorizLine((int)((x1 + 16) >> 5), (int)((x2 + 16) >> 5), iy1);
 		return;
 	}

	if (y1 < y2) {		/* down */
 		starty = iy1;
 		endy = iy2 - 1;
 		incdec = 1;
 	}
 	else {				/* up */
 		starty = iy1 - 1;
 		endy = iy2;
 		incdec = -1;
 	}

	dx = x2 - x1;
 	dy = y2 - y1;
 
	for (y = starty; y != endy + incdec; y += incdec) {
 		xx = ((((LONG)y << 5) + 16 - y1) * dx) / dy;
 		x = (int)((x1 + xx) >> 5);
 		TogglePixel(x, y);
 	}
}

static void DrawBezier(x0, x1, x2, x3, y0, y1, y2, y3, n)
LONG		x0, x1, x2, x3, y0, y1, y2, y3;
int			n;
{
	LONG    mx0, mx1, mx2, mx3, mx4, mx5;
	LONG    my0, my1, my2, my3, my4, my5;

	if (n >= nMaxSubdivDepth) {
		DrawLine(x0, y0, x3, y3);
		return;
	}

	mx0 = (x0 + x1) / 2;
	mx1 = (x1 + x2) / 2;
	mx2 = (x2 + x3) / 2;
	mx3 = (mx0 + mx1) / 2;
	mx4 = (mx1 + mx2) / 2;
	mx5 = (mx3 + mx4) / 2;
	my0 = (y0 + y1) / 2;
	my1 = (y1 + y2) / 2;
	my2 = (y2 + y3) / 2;
	my3 = (my0 + my1) / 2;
	my4 = (my1 + my2) / 2;
	my5 = (my3 + my4) / 2;
	DrawBezier(x0, mx0, mx3, mx5, y0, my0, my3, my5, n+1);
	DrawBezier(mx5, mx4, mx2, x3, my5, my4, my2, y3, n+1);
}

static SHORT GetWord()
{
	SHORT	n = (*lpCurFontData << 8) + *(lpCurFontData+1);
	lpCurFontData += 2;
	return n;
}

static short DrawGlyph()
{
	BOOL		bFirstMove;
	int			i;
	LONG		firstx, firsty, prevx, prevy, x1, y1, x2, y2, x3, y3;
	LONG		yScaledSize;

	bFirstMove = TRUE;
	firstx = prevx = 0;
	firsty = prevy = 0;

	yScaledSize = (LONG)nCharBitmapHeight << 10;

	for (i = 0; i < nCharBitmapHeight; i++)
		anCount[i] = 0;

	while (1) {
		switch (*lpCurFontData++) {
			case MOVETO:
				if (!bFirstMove)
					DrawLine(prevx, prevy, firstx, firsty);

				bFirstMove = FALSE;
				x1 = GetWord();
				y1 = GetWord();
				ScaleCoord(&x1, &y1);
				firstx = prevx = x1 - xShift;
				firsty = prevy = yScaledSize - y1 + yShift;
				break;

			case LINETO:
				x1 = GetWord();
				y1 = GetWord();
				ScaleCoord(&x1, &y1);
				x1 = x1 - xShift;
				y1 = yScaledSize - y1 + yShift;
				DrawLine(prevx, prevy, x1, y1);
				prevx = x1;
				prevy = y1;
				break;

			case CURVETO:
				x1 = GetWord();
				y1 = GetWord();
				ScaleCoord(&x1, &y1);
				x1 = x1 - xShift;
				y1 = yScaledSize - y1 + yShift;
				x2 = GetWord();
				y2 = GetWord();
				ScaleCoord(&x2, &y2);
				x2 = x2 - xShift;
				y2 = yScaledSize - y2 + yShift;
				x3 = GetWord();
				y3 = GetWord();
				ScaleCoord(&x3, &y3);
				x3 = x3 - xShift;
				y3 = yScaledSize - y3 + yShift;
				DrawBezier(prevx, x2, x3, x1, prevy, y2, y3, y1, 0);
				prevx = x1;
				prevy = y1;
				break;

			case ENDCHAR:
				goto end;

			default:
				return FALSE;
		 }
	}
end:
	DrawLine(prevx, prevy, firstx, firsty);

	return TRUE;
}


static void Sort(a, n)
SHORT a[];
int n;
{
	int		i, j;
	SHORT	t;

	for (i = 1; i < n; i++) {
		t = a[i];
		for (j = i; j > 0 && a[j-1] > t; j--) a[j] = a[j-1];
		a[j] = t;
	}
}


static void Fill()
{
	int		i, j, k;
	int		x1, x2;
	LPSTR	lpFrameBuf = lpCurFrameBuf;

	for (i = 0; i < nCharBitmapHeight; i++, lpFrameBuf += nCharXBytes) {
		Sort(alpPoint[i], anCount[i]);
		for (j = 0; j < anCount[i]; j += 2) {
			x1 = alpPoint[i][j];
			x2 = alpPoint[i][j+1];
			for (k = x1; k <= x2; k++)
				*(lpFrameBuf + (k >> 3)) |= mask[k & 7];
		}
	}
}



#include "wancode.h"

static UBCHAR KSSM2KS(ch)
UBCHAR ch;
{
	BYTE	c1, c2;
	UBCHAR	code, tmp;
	int		i, l, u, m;

	c1 = ch >> 8;
	c2 = ch & 0x00ff;

	if (c1 >= 0x84 && c1 <= 0xd3) {			/* KSSM hangul */
		l = 0;
		u = NUMHANGUL;
		while (l <= u) {
			m = (l + u) / 2;
			if (wCodeTable[m] == ch) break;
			else if (wCodeTable[m] > ch) u = m - 1;
			else l = m + 1;
		}
		if (l <= u)
			code = ((m / 94) + 0xb0) * 256 + (m % 94) + 0xa1;
		else {
			for (i = 0; i < SINGLEHAN; i++) if (ch == SingleHan[i]) break;
			if (i < SINGLEHAN)
				code = (0xa4 * 256) + i + 0xa1;
			else
				code = 0xffff;
		}
	}
	else if (c1 >= 0xd9 && c1 <= 0xde) {	/* KSSM special */
		if (c2 >= 128) c2 -= 18;
		tmp = (c1 - 0xd9) * 94 * 2 + c2 - 0x31;
		code = ((tmp / 94) + 0xa1) * 256 + (tmp % 94 + 0xa1);
	}
	else if (c1 >= 0xe0 && c1 <= 0xf9) {	/* KSSM hanja */
		if (c2 >= 128) c2 -= 18;
		tmp = (c1 - 0xe0) * 94 * 2 + c2 - 0x31;
		code = ((tmp / 94) + 0xca) * 256 + (tmp % 94 + 0xa1);
	}
	else if (c1 == 0xd4) {					/* TG special */
		return 0xffff;
	}

}

void
swap2bytes(x)
unsigned short *x;
{
	union
	{
		unsigned short s;
		unsigned char  b[2];

	} b2;
	unsigned char t;

	b2.s = *x;
	t = b2.b[0];
	b2.b[0] = b2.b[1];
	b2.b[1] = t;
	*x = b2.s;
}

void swap4bytes(x)
unsigned long *x;
{
	union
	{
		unsigned long s;
		unsigned char b[4];
	} b4;
	unsigned char t;

	b4.s = *x;
	t = b4.b[0];
	b4.b[0] = b4.b[3];
	b4.b[3] = t;
	t = b4.b[1];
	b4.b[1] = b4.b[2];
	b4.b[2] = t;
	*x = b4.s;
}
