/****************************************************************
*
*	FONTDEMO.C - Outline Font Rasterizer Demo Program
*
*  	Programmed by : Lee, Jun-Hee  (member of Technical staff, HCI)    
*  	Modified By Kim Bum Chul 1995.12.22                              
*
*	Copyright (c) 1992 by Human Computers, Inc.
*	Copyright (c) 1995 by TriGem Microsystems, Inc.
*
*****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <fcntl.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "drawchar.h"
#include "xfunc.h"


extern short SinTbl[];
void DrawBigChar();
void DrawSpiral();
void DrawWheel();
void DrawSlant();
void Wait();
void delay();

char *colors[] = {
    "White", "Black","Grey",  "Navy","Cyan", "Blue", "Violet", "Purple",
    "Pink", "Red",  "Turquoise","Magenta", "Green", "Limegreen", "Yellow",
    "Orange", "Wheat", "Brown"
};

extern Display                 *g_Disp;
extern Window                  g_RootWin, g_MainWin;
extern GC		       g_GC;

void main(ac, av)
int		ac;
char	*av[];
{
	if (!InitFontMachinery("."))
		return;
	InitPage(ac, av);

	while(1) {
		DrawBigChar();
		DrawWheel();
		DrawSlant();
		DrawSpiral();
	}
	CloseFontMachinery();
}


void DrawBigChar()
{
	int		i;
	int		x, y, r, xc, yc, color;
	static	int	size[10]={16, 24, 32, 40, 48, 80, 120, 300, 512, 1000};
	unsigned char	s[100];


	ResetCTM();

	for (i = 0; i < 10; i++) {
		ClearPage("blue");
		SetColor("yellow");
		SetFontSize(32,32);
		sprintf(s,"%dx%d", size[i], size[i]);
		DrawString(g_Disp,g_MainWin,g_GC,10, 32, s);

		SetFontSize(size[i], size[i]);
		DrawString(g_Disp,g_MainWin,g_GC,0, 40+size[i], "Ʈ");

		if (_kbhit()) {
			if (_getch() == 27) {
				CloseFontMachinery();
				exit(-1);
			}
		}

		Wait(2);
	}
}



void DrawWheel()
{
	int		i, x, y, xc, yc, n;
	int		SinVal, CosVal;

	static char *DemoStr[4] = { "","", "", "ü" };
	static	char	*aa = "";

	ClearPage("magenta");
	ResetCTM();
	SetFontSize(160, 160);
	n = 0;

	for (y = 196; y < 200; y += 96) {
		x = 0;
		for (i = 0; i < 1080; i += 5) {
			ClearPage("magenta");
			if (++n >= 100) n = 0;
			SinVal = SinTbl[i % 360];
			CosVal = SinTbl[(i + 90) % 360];
			xc = x + ((long)-60 * (CosVal - SinVal)) / 1024;
			yc = y - ((long)-60 * (SinVal + CosVal)) / 1024;
			SetColor("yellow");
			DrawString(g_Disp,g_MainWin,g_GC,xc, yc,  DemoStr[n/25]);
			x += 5;
			RotateFont(5);
			delay(50);

			if (_kbhit()) {
				if (_getch() == 27) {
					CloseFontMachinery();
					exit(-1);
				}
			}
		}
	}
	Wait(2);
}


void DrawSlant()
{
	int		i;

	ClearPage("green");
	ResetCTM();
	for (i = 0; i <= 45; i += 5) {
		ResetCTM();
		SetFontSize(160, 160);
		SlantFont(-i);
		SetColor("lightcyan");
		DrawString(g_Disp,g_MainWin,g_GC,200, 380, "ﺸũνý");
		if (_kbhit()) {
			if (_getch() == 27) {
				CloseFontMachinery();
				exit(-1);
			}
		}
		ClearPage("green");
	}
	for (i = 45; i >= -45; i -= 5) {
		ResetCTM();
		SetFontSize(160, 160);
		SlantFont(-i);
		SetColor("red");
		DrawString(g_Disp,g_MainWin,g_GC,200, 380, "ö ");
		if (_kbhit()) {
			if (_getch() == 27) {
				CloseFontMachinery();
				exit(-1);
			}
		}
		ClearPage("green");
	}
	for (i = -45; i <= 0; i += 5) {
		ResetCTM();
		SetFontSize(160, 160);
		SlantFont(-i);
		SetColor("viollet");
		DrawString(g_Disp,g_MainWin,g_GC,200, 380, "ѱƮ");
		if (_kbhit()) {
			if (_getch() == 27) {
				CloseFontMachinery();
				exit(-1);
			}
		}
	}
	Wait(2);
}


void DrawSpiral()
{
	int		i,j;
	int		x, y, r, xc, yc;

	ClearPage("blue");
	ResetCTM();
	SetFontSize(180, 180);

	xc = 320;
	yc = 240;
	for (i = 0,j=0; i < 720; i += 10,j++) {
		SetFontSize(4 + i / 6, 4 + i / 6);
		r = 4 + i / 4;
		ResetCTM();
		SetColor(colors[j%18]);	
		RotateFont(i);
		x = xc + (int)((long)r * SinTbl[(i+90) % 360] / 1024);
		y = yc - (int)((long)r * SinTbl[i % 360] / 1024);
		DrawString(g_Disp,g_MainWin,g_GC,x, y, "ﺸũνý");
		if (_kbhit()) {
			if (_getch() == 27) {
				CloseFontMachinery();
				exit(-1);
			}
		}
	}
	Wait(2);
}


static int	dummy;
static void delay(dur)
int	dur;
{
	int i,j;

	for(j=0;j<dur;j++)
		for(i=0;i<500;i++)
			dummy = i+j;
}


void Wait(sec)
int	sec;
{
	int		n = 0;
	int		i;

	sec *= 10;
	while (!_kbhit()) {
		delay(100);
		if (++n > sec) return;
	}
	
	if (_getch() == 27) {
		CloseFontMachinery();
		exit(-1);
	}
}
