#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <strings.h>

#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "xfunc.h"

void CreateMainWindow();
void HandleGeneralEvent();

Display			*g_Disp;
int			g_Screen;
Window			g_RootWin, g_MainWin;
GC			g_GC;
Visual			*g_Visual;
int			g_WinWidth, g_WinHeight;
unsigned long		g_fcol, g_bcol;

void InitPage(argc, argv)
int		argc;
char	**argv;
{
	char	*display, *geom;

	display = geom = NULL;
	/* Open the display */
	if ((g_Disp=XOpenDisplay(display)) == NULL) {
		fprintf(stderr, "\nCan't open display\n");
		exit(1);
	}

	g_Screen 	= DefaultScreen(g_Disp);
	g_RootWin	= RootWindow(g_Disp, g_Screen);
	g_GC		= DefaultGC(g_Disp, g_Screen);
	g_Visual	= DefaultVisual(g_Disp, g_Screen);
	
	g_WinWidth 	= DisplayWidth(g_Disp, g_Screen);
	g_WinHeight	= DisplayHeight(g_Disp, g_Screen);
	g_bcol = WhitePixel(g_Disp, g_Screen);
	g_fcol = BlackPixel(g_Disp, g_Screen);
	XSetForeground(g_Disp, g_GC, g_fcol);
	XSetBackground(g_Disp, g_GC, g_bcol);

	CreateMainWindow(geom, argc, argv);

	XSelectInput(g_Disp, g_MainWin, 
					ExposureMask|KeyPressMask|
					ButtonPressMask|StructureNotifyMask);

	XMapWindow(g_Disp, g_MainWin);
}



void CreateMainWindow(geom, argc, argv)
char	*geom;
int		argc;
char	**argv;
{

	XSetWindowAttributes	xswa;
	unsigned int			xswamask;
	XSizeHints				hints;
	int						i,x,y,w,h;

	x=y=w=h=1;
	i = XParseGeometry(geom, &x, &y, &w, &h);
	if (i&WidthValue) g_WinWidth = w;
	if (i&HeightValue) g_WinHeight = h;

	if (i&XValue || i&YValue) hints.flags = USPosition;
	else hints.flags = PPosition;

	hints.flags|= USSize;

	if (i&XValue && i&XNegative)
		x = XDisplayWidth(g_Disp, g_Screen)-g_WinWidth-abs(x);
	if(i&YValue && i&YNegative)
		y = XDisplayHeight(g_Disp, g_Screen)-g_WinHeight-abs(y);

	hints.x=x;			hints.y=y;
	hints.width = g_WinWidth;	hints.height = g_WinHeight;
	hints.max_width = DisplayWidth(g_Disp, g_Screen);
	hints.max_height = DisplayHeight(g_Disp, g_Screen);
	hints.flags |= PMaxSize;

	xswa.background_pixel = g_bcol;
	xswa.border_pixel = g_bcol;
	xswamask = CWBackPixel | CWBorderPixel;

	g_MainWin = XCreateWindow(g_Disp,g_RootWin,x,y,g_WinWidth,g_WinHeight,2,0,CopyFromParent,CopyFromParent,xswamask,&xswa);
	XSetStandardProperties(g_Disp,g_MainWin,"fontdemo","fontdemo",None,argv,argc,&hints);
	
	if (!g_MainWin) {
		fprintf(stderr,"\nCan't open main window\n");
		exit(1);
	}
}


void ClearPage(color) 
char *color;
{
        Colormap cmap = DefaultColormap(g_Disp, g_Screen);
        XColor col, unused;
        char buf[32];

    	if (!XAllocNamedColor(g_Disp, cmap, color, &col, &unused)) {
        	sprintf(buf, "Can't alloc %s", color);
        	return;
    	}
        XSetForeground(g_Disp, g_GC, col.pixel);

	XFillRectangle(g_Disp, g_MainWin, g_GC, 0,0, g_WinWidth,g_WinHeight);
}

void SetColor(color)
char *color;
{
        Colormap cmap = DefaultColormap(g_Disp, g_Screen);
        XColor col, unused;
        char buf[32];

    	if (!XAllocNamedColor(g_Disp, cmap, color, &col, &unused)) {
        	sprintf(buf, "Can't alloc %s", color);
        	return;
    	}
        XSetForeground(g_Disp, g_GC, col.pixel);
}

static XEvent	event;

int Xkbhit()
{

	if (XCheckWindowEvent(g_Disp, g_MainWin, 
							KeyPressMask|ButtonPressMask,&event)) {
		if (event.type == KeyPress) return 1;
		else {
			if (event.type != -1)
				HandleGeneralEvent(&event);
			return 0;
		}
	}
	else 
		return 0;
}

int Xgetch()
{
	if (event.type == KeyPress) {
		XKeyEvent	*key_event = (XKeyEvent *)&event;
		char	buf[128];
		KeySym	ks;
		XComposeStatus	status;
		XLookupString(key_event,buf,128,&ks,&status);
		event.type=-1;
		return(buf[0]);
	}
	else {
		if (event.type != -1)
			HandleGeneralEvent(&event);
		XNextEvent(g_Disp, &event);
		return(Xgetch());
	}
}


void HandleGeneralEvent(event)
XEvent	*event;
{
/**********************************************************************
	switch (event->type) {
		case ButtonPress:
			printf("Button Pressed\n");
			break;
		case KeyPress:
			printf("Key Pressed\n");
			break;
		case Expose:
			printf("Exposed\n");
			break;
		case ConfigureNotify:
			printf("ConfigureNotify\n");
			break;
		case CirculateNotify:
			printf("CirculateNotify\n");
			break;
		case MapNotify:
			printf("MapNotify\n");
			break;
		case DestroyNotify:
			printf("DestroyNotify\n");
			break;
		case GravityNotify:
			printf("GravityNotify\n");
			break;
		case ReparentNotify:
			printf("ReparentNotify\n");
			break;
		case UnmapNotify:
			printf("UnmapNotify\n");
			break;
		default:
			printf("To be processed event type=%d\n",event->type);
	}
******************************************************************/
}


int GetPageHeight()
{
	return g_WinHeight;
}
